// Star Fox 2 LZ Big-Endian Decoder

#include <stdio.h>
#include <memory.h>

#define u32 unsigned int
#define u8  unsigned char
#define s8    signed char

FILE *in,*out,*check=0;

//#define TRACE

// decoder-related
u8 buffer[0x10000];
int counter;
int dest = 0xffff;

// file-related
u8 file[0x10000];
int file_ptr = 0xffff;
u32 file_buf = 0;
int mask = 0;
bool first = 1;

int READ_BYTE()
{
	return file[file_ptr--];
}

int READ_WORD()
{
	int value = READ_BYTE();
	value |= (READ_BYTE())<<8;
	return value;
}

int READ_DWORD()
{
	int value = READ_WORD();
	value |= (READ_WORD())<<16;
	return value;
}

int GetBits(int bits)
{
	int value=0;

	// only use 'x' bits
	while(bits)
	{
		// append bits
		value <<= 1;
		value |= file_buf&1;

		// subtract usage
		file_buf >>= 1;
		mask--;

		bits--;

		// refresh barrel
		if(!mask || (first && !file_buf))
		{
			// compensate
			if(first)
			{
				bits++;
				value >>= 1;
			}

			file_buf = READ_BYTE();
			mask = 8;
			first = 0;
		}
	}

	return value;
}

void WRITE_BYTE(int out, FILE *fp)
{
	fputc(out&0xff,fp);
}

void Decode_Free()
{
	int run;

	// Grab decode method
	run = GetBits(3);

	// extended mode
	if(run == 7)
	{
		if(GetBits(1) == 0)
		{
			run = GetBits(4) + 7;
		}
		else
		{
			run = GetBits(10);
			if(run == 0)
			{
				run = GetBits(18);
			}
		}
	}

#ifdef TRACE
	printf("[%04X] Free: %d\n",0x10000-(dest+1),run);
#endif

	// copy 'x' free bytes
	while(run)
	{
		int byte = GetBits(8);

		// update LZ buffer
		buffer[dest--] = byte;
		counter--;

		run--;
	}
}

void Decode_LZ()
{
	int run;
	int delta;
	int method;

	// Grab decode method
	method = GetBits(2);

	if(method == 0)
	{
		run = 2;
		delta = GetBits(8);
	}
	else if(method == 1)
	{
		run = 3;
		if(GetBits(1) == 1)
			delta = GetBits(8);
		else
			delta = GetBits(14);
	}
	else
	{
		if(method == 2)
			run = 4;
		else
		{
			method = GetBits(2);

			if(method == 3)
				run = GetBits(8);
			else if(method == 2)
				run = GetBits(2)+7;
			else
				run = method+5;
		}

		if(GetBits(1) == 0)
			delta = GetBits(16);
		else if(GetBits(1) == 1)
			delta = GetBits(8);
		else
			delta = GetBits(12);
	}

#ifdef TRACE
	printf("[%04X] LZ:   %d, %04X\n",0x10000-(dest+1),run,delta);
#endif

	// copy 'x' LZ bytes
	while(run)
	{
		int byte = buffer[dest + delta];

		// update LZ buffer
		buffer[dest--] = byte;
		counter--;

		run--;
	}
}

void Decode()
{
	int unknown;

	// Init
	counter = READ_WORD();
	unknown = READ_WORD();
	file_buf = READ_DWORD();
	mask = 32;

	// Decoder
	while(1)
	{
		Decode_Free();
		if(counter) Decode_LZ();
		else break;
	}

	for(int lcv=dest+1; lcv<0x10000; lcv++)
	{
		WRITE_BYTE(buffer[lcv],out);
	}
}

int main(int argc, char* argv[])
{
	int start;

	if(argc<3) return 0;

	in = fopen(argv[1],"rb");
	out = fopen(argv[2],"wb");

	// Image file
	if(argc==4)
	{
		check = fopen("check.bin","wb");

		// buffer file data (reverse storage)
		sscanf(argv[3],"%08X",&start);
		if(start>=0x10000)
		{
			fseek(in,start-0x10000,SEEK_SET);
			fread(file,0x10000,1,in);
		}
		else
		{
			fread(file,0x10000,1,in);
			file_ptr = start-1;
		}
	}
	// Raw file
	else
	{
		fread(file,0x10000,1,in);
		file_ptr = ftell(in)-1;
	}

	Decode();

	fclose(in);
	fclose(out);

	if(check)
	{
		for(int lcv=file_ptr+1; lcv<0x10000; lcv++)
		{
			WRITE_BYTE(file[lcv],check);
		}
		fclose(check);
	}

	return 0;
}