// Star Fox 2: Font inserter/extractor

#include <stdio.h>
#include <string.h>

typedef unsigned char u8;

FILE *image,*fp,*width;
u8 matrix[16][203];	// 8x12 font grid

void Extract_Font()
{
	int font;
	int total=0;

	fseek(image,0x6E45A,SEEK_SET);

	////////////////////////////////////////////////
	// Width table

	// note: some large "characters" are broken up into
	//       smaller pieces

	for(font=0; font<209; font++)
	{
		int data;

		// be careful of optimization here
		data  = fgetc(image);
		data |= (fgetc(image)<<8);

		// starting bit position: data & 0x07
		// index into font: (data>>3) & 0xff
		int font_width = data>>11;

		fputc(font_width,width);
		total += font_width;
	}

	// compute how many bytes this should absorb
	if(total & 7) total+=8;
	total >>= 3;

	////////////////////////////////////////////////
	// 1-bpp FX linear ==> 1-bpp SNES tile (top-down format)

	// Cache font data
	fread(matrix,total*12,1,image);

	for(font=0; font<total; font++)
	{
		for(int row=0; row<16; row++)
		{
			int byte;
			
			// non-font or real data
			byte = (row>=12) ? 0 : matrix[row][font];

			fputc(byte,fp);
		}
	}
}

void Insert_Font()
{
	int total=0;
	int font;

	fseek(image,0x6E45A,SEEK_SET);

	////////////////////////////////////////////////
	// Width table

	int bit_pos=0;
	int index=0;

	for(font=0; font<209; font++)
	{
		int font_width = fgetc(width);
		int data = (font_width<<11) | (index<<3) | (bit_pos);

		fputc(data&0xff,image);
		fputc(data>>8,image);

		// update font buffer
		bit_pos += font_width;
		while(bit_pos>7)
		{
			// next boundary
			bit_pos -= 8;
			index++;
		}
		total += font_width;
	}

	// compute how many bytes this should absorb
	if(total & 7) total+=8;
	total >>= 3;

	////////////////////////////////////////////////
	// 1-bpp SNES tile (top-down format) ==> 1-bpp FX linear

	for(font=0; font<total; font++)
	{
		for(int row=0; row<16; row++)
		{
			unsigned char byte = fgetc(fp);

			// only valid font data
			if(row<12) matrix[row][font] = byte;
		}
	}

	// plug matrix back into image
	fwrite(matrix,1,total*12,image);
}

int main(int argc, char**argv)
{
	if(strcmp(argv[1],"extract")==0)
	{
		image = fopen(argv[2],"rb");
		fp = fopen(argv[3],"wb");
		width = fopen(argv[4],"wb");

		Extract_Font();

		fclose(image);
		fclose(fp);
		fclose(width);
	}
	else if(strcmp(argv[1],"insert")==0)
	{
		image = fopen(argv[2],"r+b");
		fp = fopen(argv[3],"rb");
		width = fopen(argv[4],"rb");

		Insert_Font();

		fclose(image);
		fclose(fp);
		fclose(width);
	}

	return 0;
}